@@ -5,12 +5,12 @@
 # Do not include any function in here that does not correspond to a COMMAND
 
 import os
-import sys
 import inspect
 import tempfile
 import subprocess
 import urllib.request
 import webbrowser
+import shutil
 
 import apt
 
@@ -34,7 +34,8 @@ def addrepo(args):
     An example that shows how to add the daily builds of
     Google's Chromium browser:
 
-    $ wajig addrepo ppa:chromium-daily"""
+    $ wajig addrepo ppa:chromium-daily
+    """
     util.requires_package("add-apt-repository")
     perform.execute("/usr/bin/add-apt-repository " + args.ppa, root=True)
 
@@ -46,8 +47,9 @@ def aptlog(args):
 
 def autoalts(args):
     """Mark the Alternative to be auto-set (using set priorities)"""
-    perform.execute("/usr/sbin/update-alternatives --auto " + args.alternative,
-                     root=True)
+    perform.execute(
+        "/usr/bin/update-alternatives --auto " + args.alternative, root=True
+    )
 
 
 def autodownload(args):
@@ -59,7 +61,7 @@ def autodownload(args):
     if not args.simulate:
         upgradable_packages = util.upgradable()
         if upgradable_packages:
-            util.do_describe(upgradable_packages, verbose)
+            util.do_describe(upgradable_packages, args.verbose)
         else:
             print("no upgradable packages")
         util.show_package_versions()
@@ -77,7 +79,9 @@ def autoremove(args):
 
 def build(args):
     """Get source packages, unpack them, and build binary packages from them.
-    This also installs the needed build-dependencies if needed."""
+
+    Note: This also installs the needed build-dependencies if needed
+    """
     util.requires_package("sudo")
     # First make sure dependencies are met
     if not builddeps(args):
@@ -104,14 +108,15 @@ def changelog(args):
     network off:
          changelog - if there's newer entries, mention failure to retrieve
       -v changelog - if there's newer entries, mention failure to retrieve, and
-                     proceed to display complete local changelog"""
+                     proceed to display complete local changelog
+    """
 
     package = util.package_exists(apt.Cache(), args.package)
     changelog = "{:=^79}\n".format(" {} ".format(args.package))  # header
 
     try:
         changelog += package.get_changelog()
-    except AttributeError as e:
+    except AttributeError:
         # This is caught so as to avoid an ugly python-apt trace; it's a bug
         # that surfaces when:
         # 1. The package is not available in the default Debian suite
@@ -145,7 +150,10 @@ def changelog(args):
             perform.execute(command)
         with open(tmp) as f:
             for line in f:
-                print(line, end="")
+                try:
+                    print(line, end="")
+                except BrokenPipeError:
+                    return
 
 
 def clean(args):
@@ -173,7 +181,8 @@ def dependents(args):
     * Recommends
     * Suggests
     * Replaces
-    * Enhances"""
+    * Enhances
+    """
 
     DEPENDENCY_TYPES = [
         "Depends",
@@ -185,7 +194,7 @@ def dependents(args):
 
     cache = apt.cache.Cache()
     package = util.package_exists(cache, args.package)
-    dependents = { name : [] for name in DEPENDENCY_TYPES }
+    dependents = {name : [] for name in DEPENDENCY_TYPES}
 
     for key in cache.keys():
         other_package = cache[key]
@@ -196,8 +205,9 @@ def dependents(args):
 
     for dependency_type, specific_dependents in dependents.items():
         if specific_dependents:
-            output = dependency_type.upper(), " ".join(specific_dependents)
-            print("{}: {}".format(*output))
+            print("{}: {}".format(
+                dependency_type.upper(), " ".join(specific_dependents)
+            ))
 
 
 def describe(args):
@@ -209,7 +219,8 @@ def describe(args):
 def describenew(args):
     """Display one-line descriptions of newly-available packages
 
-    This produces the same output as 'wajig new'"""
+    This produces the same output as 'wajig new'
+    """
     util.newly_available()
 
 
@@ -228,8 +239,9 @@ def distupgrade(args):
         util.requires_package("dpkg-repack")
         util.requires_package("fakeroot")
         util.backup_before_upgrade(packages, distupgrade=True)
-    cmd = "/usr/bin/apt-get --show-upgraded {} {} {} ".format(args.local, args.yes,
-                                                              args.noauth)
+    cmd = "/usr/bin/apt-get --show-upgraded {} {} {} ".format(
+        args.local, args.yes, args.noauth
+    )
     if args.dist:
         cmd += "--target-release " + args.dist + " "
     cmd += "dist-upgrade"
@@ -247,6 +259,7 @@ def download(args):
 
 def editsources(args):
     """Edit list of Debian repository locations for packages"""
+    # stolen from http://selenic.com/hg/file/889789a2ca9f/mercurial/ui.py#l836
     editor = (
         os.environ.get('VISUAL') or
         os.environ.get('EDITOR') or
@@ -282,9 +295,10 @@ def fixmissing(args):
 def force(args):
     """Install packages and ignore file overwrites and depends
 
-    note: This is useful when there is a conflict of the same file from
+    Note: This is useful when there is a conflict of the same file from
           multiple packages or when a dependency is not installed for
-          whatever reason"""
+          whatever reason
+    """
 
     command = "/usr/bin/dpkg --install --force overwrite --force depends "
     archives = "/var/cache/apt/archives/"
@@ -308,15 +322,17 @@ def force(args):
         for package in args.packages:
             # Identify the latest version of the package available in
             # the download archive, if there is any there.
-            lscmd = "ls " + archives
-            lscmd += " | grep -E '^" + package + "_' | sort -k 1b,1 | tail -n -1"
+            lscmd = "ls {} | grep -E '^{}_' | sort -k 1b,1 | tail -n -1"
+            lscmd = lscmd.format(archives, package)
             matches = perform.execute(lscmd, pipe=True)
             debpkg = matches.readline().strip()
 
             if not debpkg:
-                dlcmd = "apt-get --quiet=2 --reinstall --download-only "
-                dlcmd += "install '" + package + "'"
-                perform.execute(dlcmd, root=1)
+                dlcmd = (
+                    "apt-get --quiet=2 --reinstall --download-only "
+                    "install '{}'"
+                ).format(package)
+                perform.execute(dlcmd, root=True)
                 matches = perform.execute(lscmd, pipe=True)
                 debpkg = matches.readline().strip()
 
@@ -331,8 +347,8 @@ def hold(args):
     for package in args.packages:
         # The dpkg needs sudo but not the echo.
         # Do all of it as root then!
-        command = "/bin/echo \"" + package + " hold\" | /usr/bin/dpkg --set-selections"
-        perform.execute(command, root=True)
+        command = '/bin/echo "{} hold" | /usr/bin/dpkg --set-selections'
+        perform.execute(command.format(package), root=True)
     print("The following packages are on hold:")
     perform.execute("dpkg --get-selections | grep -E 'hold$' | cut -f1")
 
@@ -360,12 +376,15 @@ def install(args):
     $ wajig install a b_1.0_all.deb http://example.com/c_1.0_all.deb
 
     Assuming there's no errors, the command will install 3 packages named
-    'a', 'b', and 'c''"""
+    'a', 'b', and 'c''
+    """
 
     packages = util.consolidate_package_names(args)
 
-    online_files = [package for package in packages
-                            if package.startswith(("http://", "ftp://"))]
+    online_files = [
+        package for package in packages if
+        package.startswith(("http://", "ftp://"))
+    ]
     deb_files = list()
     for package in online_files:
         if not package.endswith(".deb"):
@@ -381,11 +400,12 @@ def install(args):
                 f.write(response.read())
             deb_files.append(filename)
 
-    deb_files.extend([package for package in packages
-                            if package.endswith(".deb")
-                            and os.path.exists(package)])
+    deb_files.extend([
+        package for package in packages if package.endswith(".deb")
+        and os.path.exists(package)
+    ])
     if deb_files:
-        debfile.install(deb_files, args)
+        debfile.install(deb_files)
 
     packages = packages.difference(online_files, deb_files)
     if packages:
@@ -415,6 +435,7 @@ def installsuggested(args):
 
 def integrity(args):
     """Check the integrity of installed packages (through checksums)"""
+    util.requires_package("debsums")
     perform.execute("debsums --all --silent")
 
 
@@ -477,22 +498,25 @@ def listcommands(args):
 
 def listdaemons(args):
     """List the daemons that wajig can start, stop, restart, or reload"""
-    command = ("printf 'Found %d daemons in /etc/init.d.\n\n' "
-               "$(ls /etc/init.d/ | "
-               "grep -E -v '(~$|README|-(old|dist)|\.[0-9]*$)' | wc -l)")
-    perform.execute(command)
-    command = ("ls /etc/init.d/ | "
-               "grep -E -v '(~$|README|-(old|dist)|\.[0-9]*$)' |"
-               "pr --columns=3 --omit-header")
-    perform.execute(command)
+    util.requires_package("chkconfig")
+    perform.execute("chkconfig")
 
 
 def listfiles(args):
     """List the files that are supplied by the named package"""
-    if args.package.endswith("deb"):
+    if args.package.endswith(".deb"):
         perform.execute("dpkg --contents " + args.package)
+        return
+    try:
+        output = perform.execute(
+            "dpkg --listfiles " + args.package, getoutput=True
+        )
+    except subprocess.CalledProcessError:
+        if shutil.which("apt-file"):
+            perform.execute("apt-file list --regexp ^{}$".format(args.package))
     else:
-        perform.execute("dpkg --listfiles " + args.package)
+        for line in output.decode().strip().split('\n'):
+            print(line)
 
 
 def listhold(args):
@@ -554,12 +578,12 @@ def listscripts(args):
 def listsection(args):
     """List packages that belong to a specific section
 
-    note: Use the LISTSECTIONS command for a list of Debian Sections"""
-    section = args.section
+    Note: Use the LISTSECTIONS command for a list of Debian Sections
+    """
     cache = apt.cache.Cache()
     for package in cache.keys():
         package = cache[package]
-        if(package.section == args.section):
+        if package.section == args.section:
             print(package.name)
 
 
@@ -585,16 +609,12 @@ def liststatus(args):
     perform.execute(command)
 
 
-def localdistupgrade(args):
-    """Dist-upgrade using only packages that are already downloaded"""
-    command = ("/usr/bin/apt-get --no-download --ignore-missing --show-upgraded "
-               "dist-upgrade")
-    perform.execute(command, root=True, log=True)
-
-
 def localupgrade(args):
     """Upgrade using only packages that are already downloaded"""
-    command = "/usr/bin/apt-get --no-download --ignore-missing --show-upgraded upgrade"
+    command = (
+        "/usr/bin/apt-get --no-download --ignore-missing "
+        "--show-upgraded upgrade"
+    )
     perform.execute(command, root=True, log=True)
 
 
@@ -605,7 +625,8 @@ def madison(args):
 
 
 def move(args):
-    """Move packages in the download cache to a local Debian mirror (apt-move)"""
+    """Move packages in the download cache to a local Debian mirror"""
+    util.requires_package("apt-move")
     perform.execute("/usr/bin/apt-move update", root=True)
 
 
@@ -617,7 +638,8 @@ def new(args):
 def newdetail(args):
     """Display detailed descriptions of newly-available packages
 
-    This produces the same output as 'wajig new --verbose'"""
+    This produces the same output as 'wajig new --verbose'
+    """
     util.newly_available(verbose=True)
 
 
@@ -669,8 +691,11 @@ def purgeorphans(args):
 def purgeremoved(args):
     """Purge all packages marked as deinstall"""
     packages = ""
-    cmd = ("dpkg-query --show --showformat='${Package}:${Architecture}\t${Status}\n' | "
-           "grep -E \"deinstall ok config-files\" | cut -f 1 ")
+    cmd = (
+        "dpkg-query --show "
+        "--showformat='${Package}:${Architecture}\t${Status}\n' | "
+        "grep -E \"deinstall ok config-files\" | cut -f 1 "
+    )
     packages = perform.execute(cmd, pipe=True)
     if packages:
         packages = " ".join(packages)
@@ -717,8 +742,8 @@ def recdownload(args):
         print(package, end=' ')
     print()
 
-    command = "/usr/bin/apt-get --download-only --reinstall -u install " + args.noauth
-    command += " ".join(package_names)
+    command = "/usr/bin/apt-get --download-only --reinstall -u install {} {}"
+    command = command.format(args.noauth, " ".join(package_names))
     perform.execute(command, root=True)
 
 
@@ -730,16 +755,17 @@ def reconfigure(args):
 
 def recommended(args):
     """Display packages installed as Recommends and have no dependents"""
-    command = ("aptitude search '"
-              "?and( ?automatic(?reverse-recommends(?installed)), "
-              "?not(?automatic(?reverse-depends(?installed))) )'")
-    perform.execute(command)
+    perform.execute(
+        "aptitude search "
+        "'?and( ?automatic(?reverse-recommends(?installed)), "
+        "?not(?automatic(?reverse-depends(?installed))) )'"
+    )
 
 
 def reinstall(args):
     """Reinstall the given packages"""
-    command = "/usr/bin/apt-get install --reinstall {} {} " + " ".join(args.packages)
-    command = command.format(args.noauth, args.yes)
+    command = "/usr/bin/apt-get install --reinstall {} {} {}"
+    command = command.format(args.noauth, args.yes, " ".join(args.packages))
     perform.execute(command, root=True, log=True)
 
 
@@ -813,7 +839,7 @@ def search(args):
     balazar - adventure/action game Balazar -- Arkanae II, reforged scepters
     compizconfig-settings-manager - Compizconfig Settings Manager
     ...
-"""
+    """
     if len(args.patterns) == 1 and '::' in args.patterns[0]:
         util.requires_package('debtags')
         command = 'debtags search ' + args.patterns[0]
@@ -862,10 +888,10 @@ def stop(args):
 
 def sizes(args):
     """Display installed sizes of given packages
-    $ wajig sizes [<package name(s)>]
 
-    Display installed sizes of all packages
-    $ wajig sizes"""
+    To display sizes of all packages, do not use any argument:
+    $ wajig sizes
+    """
     util.sizes(args.packages)
 
 
@@ -924,7 +950,7 @@ def unhold(args):
         # The dpkg needs sudo but not the echo.
         # Do all of it as root then.
         command = "echo \"" + package + " install\" | dpkg --set-selections"
-        perform.execute(command, root=1)
+        perform.execute(command, root=True)
     print("The following packages are still on hold:")
     perform.execute("dpkg --get-selections | grep -E 'hold$' | cut -f1")
 
@@ -948,7 +974,7 @@ def update(args):
 
 def updatealternatives(args):
     """Update default alternative for things like x-window-manager"""
-    command = "/usr/sbin/update-alternatives --config " + args.alternative
+    command = "/usr/bin/update-alternatives --config " + args.alternative
     perform.execute(command, root=True)
 
 
@@ -977,7 +1003,9 @@ def upgrade(args):
             util.requires_package("dpkg-repack")
             util.requires_package("fakeroot")
             util.backup_before_upgrade(packages)
-        command = "/usr/bin/apt-get {} {} {} --show-upgraded --with-new-pkgs upgrade"
+        command = (
+            "/usr/bin/apt-get {} {} {} --show-upgraded --with-new-pkgs upgrade"
+        )
         command = command.format(args.local, args.yes, args.noauth)
         perform.execute(command, root=True, log=True)
     else:
@@ -992,8 +1020,10 @@ def upgradesecurity(args):
     sources_file.write("deb http://security.debian.org/ " +\
                        "testing/updates main contrib non-free\n")
     sources_file.close()
-    command = ("/usr/bin/apt-get --no-list-cleanup --option Dir::Etc::SourceList="
-               "{} update")
+    command = (
+        "/usr/bin/apt-get --no-list-cleanup --option Dir::Etc::SourceList="
+        "{} update"
+    )
     command = command.format(sources_list)
     perform.execute(command, root=True)
     command = "/usr/bin/apt-get --option Dir::Etc::SourceList={} upgrade"
@@ -1022,15 +1052,38 @@ def versions(args):
 def whichpackage(args):
     """Search for files matching a given pattern within packages
 
-    Note: if no match is found, the apt-file repository is checked"""
+    Note: also searches files for uninstalled packages if apt-file is
+    installed
+    """
     try:
-        out = perform.execute("dpkg --search " + args.pattern, getoutput=True)
+        output = perform.execute(
+            "dpkg --search " + args.pattern, getoutput=True
+        ).decode()
+    # 'dpkg --search' returns a failing error code if it does not find matches
     except subprocess.CalledProcessError:
-        util.requires_package("apt-file")
-        perform.execute("apt-file search " + args.pattern)
+        installed_matches = []
     else:
-        try:
-            print(out.decode().strip())
-        # will get here when on --simulate mode
-        except AttributeError:
-            pass
+        installed_matches = output.strip().split('\n')
+        header = "INSTALLED MATCHES (x{})".format(len(installed_matches))
+        print(header)
+        print('-' * len(header))
+        for line in installed_matches:
+            print(line)
+        print()
+    if shutil.which("apt-file"):
+        output = perform.execute(
+            "apt-file search " + args.pattern, getoutput=True
+        ).decode()
+        all_matches = output.strip()
+        if all_matches:
+            all_matches = all_matches.split('\n')
+            uninstalled_matches = set(all_matches) - set(installed_matches)
+            header = "UNINSTALLED MATCHES (x{})".format(
+                len(uninstalled_matches)
+            )
+            print(header)
+            print('-' * len(header))
+            for line in uninstalled_matches:
+                print(line)
+    else:
+        print("NOTE: install apt-file in order to display uninstalled matches")
